// Copyright 2021 The IREE Authors
//
// Licensed under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

#include "iree/compiler/Bindings/Native/Transforms/Passes.h"
#include "iree/compiler/Dialect/HAL/IR/HALDialect.h"
#include "iree/compiler/Dialect/HAL/IR/HALOps.h"
#include "llvm/ADT/STLExtras.h"
#include "mlir/Dialect/Affine/Utils.h"
#include "mlir/Dialect/Func/IR/FuncOps.h"
#include "mlir/Dialect/Tensor/IR/Tensor.h"
#include "mlir/IR/Attributes.h"
#include "mlir/IR/MLIRContext.h"
#include "mlir/IR/SymbolTable.h"
#include "mlir/Pass/Pass.h"
#include "mlir/Pass/PassRegistry.h"
#include "mlir/Support/LLVM.h"
#include "mlir/Support/LogicalResult.h"

namespace mlir {
namespace iree_compiler {
namespace IREE {
namespace ABI {

// Returns the invocation model specified on |op| or the |defaultModel|.
static IREE::ABI::InvocationModel getInvocationModel(
    Operation *op, IREE::ABI::InvocationModel defaultModel) {
  auto modelAttr = op->getAttrOfType<StringAttr>("iree.abi.model");
  if (!modelAttr) return defaultModel;
  if (modelAttr == "coarse-fences") {
    return IREE::ABI::InvocationModel::CoarseFences;
  } else {
    return IREE::ABI::InvocationModel::Sync;
  }
}

// Maps a source type to the native ABI type.
static Type mapToABIType(Type type) {
  if (type.isa<TensorType>()) {
    return IREE::HAL::BufferViewType::get(type.getContext());
  }
  return type;
}

// Creates the corresponding wrapper function for the given import function.
static func::FuncOp createImportWrapperFunc(
    IREE::ABI::InvocationModel invocationModel, func::FuncOp importOp,
    FunctionType oldImportType, FunctionType newImportType,
    StringRef privateName) {
  // Create the internal wrapper function with the original import signature.
  auto wrapperOp =
      func::FuncOp::create(importOp.getLoc(), privateName, oldImportType);
  wrapperOp.setPrivate();

  // Copy arg/result attrs from the import op to the wrapper function.
  // We may want to remove them from the import but would need to filter.
  SmallVector<DictionaryAttr, 4> argAttrDict;
  importOp.getAllArgAttrs(argAttrDict);
  wrapperOp.setAllArgAttrs(argAttrDict);
  SmallVector<DictionaryAttr, 4> resultAttrDict;
  importOp.getAllResultAttrs(resultAttrDict);
  wrapperOp.setAllResultAttrs(resultAttrDict);
  switch (invocationModel) {
    default:
    case IREE::ABI::InvocationModel::Sync:
      break;
    case IREE::ABI::InvocationModel::CoarseFences:
      argAttrDict.push_back(nullptr);  // wait
      argAttrDict.push_back(nullptr);  // signal
      break;
  }

  // Update the import type and propagate back the attributes we may have
  // modified above.
  importOp.setType(newImportType);
  importOp.setAllArgAttrs(argAttrDict);
  importOp.setAllResultAttrs(resultAttrDict);

  auto *entryBlock = wrapperOp.addEntryBlock();
  auto entryBuilder = OpBuilder::atBlockBegin(entryBlock);

  // Gather tensor arguments we may need to handle specially.
  SmallVector<Value> entryArgs = {entryBlock->getArguments().begin(),
                                  entryBlock->getArguments().end()};
  SmallVector<size_t> tensorArgIndices;
  SmallVector<Value> tensorArgs;
  for (auto [argIndex, arg] : llvm::enumerate(entryArgs)) {
    auto oldType = oldImportType.getInput(argIndex);
    if (oldType.isa<TensorType>()) {
      tensorArgIndices.push_back(argIndex);
      tensorArgs.push_back(arg);
    }
  }

  // Side-effecting imports need to have a host-wait inserted on them today.
  // We could add more configuration options here but for now we require that
  // users mark their functions 'nosideeffects' to avoid the host wait.
  const bool hasSideEffects = !importOp->hasAttr("nosideeffects");

  // When running async we insert a barrier on tensor arguments and attach that
  // to the fence we pass to the import for waiting. We'll also allocate the
  // signal fence that the import must signal when the returned tensors are
  // ready.
  Value waitFence;
  Value signalFence;
  switch (invocationModel) {
    default:
    case IREE::ABI::InvocationModel::Sync:
      // No fences.
      break;
    case IREE::ABI::InvocationModel::CoarseFences: {
      // HACK: this is relying on the fact that there's only one HAL device.
      // We should instead have a way of creating fences on the device that
      // is used to produce the tensors we're wrapping.
      auto device =
          entryBuilder.create<IREE::HAL::ExSharedDeviceOp>(importOp.getLoc());

      // When exporting a fence we need to put a barrier between the rest of the
      // program and the tensors consumed by the import.
      if (tensorArgs.empty()) {
        // No tensors passed to the import - pass in an immediate signal.
        waitFence = entryBuilder.create<IREE::Util::NullOp>(
            importOp.getLoc(), entryBuilder.getType<IREE::HAL::FenceType>());
      } else {
        waitFence = entryBuilder.create<IREE::HAL::FenceCreateOp>(
            importOp.getLoc(), entryBuilder.getType<IREE::HAL::FenceType>(),
            device, IREE::HAL::FenceFlagBitfield::None);
        auto barrierOp = entryBuilder.create<IREE::HAL::TensorBarrierOp>(
            importOp.getLoc(), tensorArgs, waitFence);
        for (auto [argIndex, readyArg] :
             llvm::zip(tensorArgIndices, barrierOp.getResults())) {
          entryArgs[argIndex] = readyArg;
        }
      }

      // When the import produces resources we need to pass in a fence it can
      // signal when execution is ready.
      // TODO(benvanik): always pass in a signal fence? could be useful if we
      // want to allow for async work using fences that's not device-related.
      const bool haveTensorResults =
          llvm::any_of(oldImportType.getResults(),
                       [](Type type) { return type.isa<TensorType>(); });
      if (!haveTensorResults && !hasSideEffects) {
        // No tensors returned from import - pass in an immediate signal.
        signalFence = entryBuilder.create<IREE::Util::NullOp>(
            importOp.getLoc(), entryBuilder.getType<IREE::HAL::FenceType>());
      } else {
        signalFence = entryBuilder.create<IREE::HAL::FenceCreateOp>(
            importOp.getLoc(), entryBuilder.getType<IREE::HAL::FenceType>(),
            device, IREE::HAL::FenceFlagBitfield::None);
      }
      break;
    }
  }

  // Marshal arguments.
  SmallVector<Value> arguments;
  for (auto [argIndex, arg] : llvm::enumerate(entryArgs)) {
    auto oldType = oldImportType.getInput(argIndex);
    auto newType = newImportType.getInput(argIndex);
    if (oldType.isa<TensorType>()) {
      // This is where we could perform type casting or in-place storage binding
      // if the user had any attrs specifying it.
      // NOTE: we insert a barrier on this above if needed so that the wait
      // fence will be signaled when the tensor is ready for consumption by the
      // import.
      auto argLoc = arg.getLoc();
      auto exportOp =
          entryBuilder.create<IREE::HAL::TensorExportOp>(argLoc, newType, arg);
      arguments.push_back(exportOp.getTarget());
    } else {
      arguments.push_back(arg);
    }
  }
  if (waitFence) arguments.push_back(waitFence);
  if (signalFence) arguments.push_back(signalFence);

  // Make the call with the updated types.
  auto callOp =
      entryBuilder.create<func::CallOp>(importOp.getLoc(), importOp, arguments);

  // If the call has side-effects then we need to wait on its signal fence on
  // the host. This is because they may have launched a thread of their own to
  // perform work that we can't track.
  if (hasSideEffects && signalFence) {
    auto timeoutMillis =
        entryBuilder.create<arith::ConstantIntOp>(importOp.getLoc(), -1, 32);
    entryBuilder.create<IREE::HAL::FenceAwaitOp>(importOp.getLoc(),
                                                 entryBuilder.getI32Type(),
                                                 timeoutMillis, signalFence);
  }

  // Marshal results.
  SmallVector<Value> results;
  for (auto [resultIndex, result] : llvm::enumerate(callOp.getResults())) {
    auto oldType = oldImportType.getResult(resultIndex);
    if (oldType.isa<TensorType>()) {
      // NOTE: we set the import pending on the signal fence from the import
      // indicating when the returned tensor is ready for consumption by the
      // program.
      results.push_back(entryBuilder.create<IREE::HAL::TensorImportOp>(
          importOp.getLoc(), oldType, result, signalFence));
    } else {
      results.push_back(result);
    }
  }

  entryBuilder.create<func::ReturnOp>(importOp.getLoc(), results);
  return wrapperOp;
}

// Updates |importOp| to use the native ABI and creates a wrapper function that
// preserves the original behavior. All callers will be updated to point at the
// new wrapper function.
static LogicalResult wrapImportFunc(IREE::ABI::InvocationModel invocationModel,
                                    mlir::ModuleOp moduleOp,
                                    func::FuncOp importOp,
                                    SymbolTable &symbolTable) {
  // Replace all existing calls to the import to instead call the wrapper.
  auto publicName = importOp.getName().str();
  auto privateName = "_" + publicName;
  auto privateNameAttr =
      mlir::StringAttr::get(importOp.getContext(), privateName);
  if (failed(symbolTable.replaceAllSymbolUses(importOp, privateNameAttr,
                                              moduleOp))) {
    return importOp.emitError() << "unknown symbol table op encountered; "
                                   "cannot fix up symbol names";
  }

  // Convert import signature types to those required by the binding ABI.
  auto oldImportType = importOp.getFunctionType();
  SmallVector<Type> inputTypes;
  for (auto oldType : oldImportType.getInputs()) {
    inputTypes.push_back(mapToABIType(oldType));
  }
  auto fenceType = IREE::HAL::FenceType::get(importOp.getContext());
  switch (invocationModel) {
    default:
    case IREE::ABI::InvocationModel::Sync:
      break;
    case IREE::ABI::InvocationModel::CoarseFences:
      inputTypes.push_back(fenceType);  // wait
      inputTypes.push_back(fenceType);  // signal
      break;
  }
  SmallVector<Type> resultTypes;
  for (auto oldType : oldImportType.getResults()) {
    resultTypes.push_back(mapToABIType(oldType));
  }
  auto newImportType =
      FunctionType::get(importOp.getContext(), inputTypes, resultTypes);

  // Create the wrapper function that matches the original internal types but
  // calls out to the updated import using ABI types.
  auto wrapperOp = createImportWrapperFunc(
      invocationModel, importOp, oldImportType, newImportType, privateName);
  if (!wrapperOp) return failure();
  moduleOp.insert(++Block::iterator(importOp), wrapperOp);

  // Update the import to the new type and mark it as being converted so we
  // don't try to convert it again.
  importOp->setAttr("iree.abi.stub", UnitAttr::get(importOp.getContext()));

  return success();
}

// Populates attributes on |wrapperOp| to support runtime reflection.
// These are attached to the exported function and can be queried at runtime
// with iree_vm_function_lookup_attr_by_name.
static void populateReflectionAttrs(IREE::ABI::InvocationModel invocationModel,
                                    func::FuncOp exportOp,
                                    func::FuncOp wrapperOp) {
  auto *context = exportOp.getContext();
  SmallVector<NamedAttribute, 4> attrs;

  if (auto abiAttr = exportOp->getAttr("iree.abi")) {
    attrs.emplace_back(StringAttr::get(context, "iree.abi"), abiAttr);
  }

  switch (invocationModel) {
    default:
    case IREE::ABI::InvocationModel::Sync:
      break;
    case IREE::ABI::InvocationModel::CoarseFences:
      attrs.emplace_back(StringAttr::get(context, "iree.abi.model"),
                         StringAttr::get(context, "coarse-fences"));
      break;
  }

  if (!attrs.empty()) {
    auto reflectionAttr = DictionaryAttr::get(context, attrs);
    wrapperOp->setAttr("iree.reflection", reflectionAttr);
  }
}

// Creates the corresponding wrapper function for the given export function.
static func::FuncOp createExportWrapperFunc(
    IREE::ABI::InvocationModel invocationModel, func::FuncOp exportOp,
    StringRef publicName) {
  // Copy arg/result attrs from the export op to the wrapper function.
  // We may want to remove them from the export but would need to filter.
  SmallVector<DictionaryAttr, 4> argAttrDict;
  exportOp.getAllArgAttrs(argAttrDict);
  SmallVector<DictionaryAttr, 4> resultAttrDict;
  exportOp.getAllResultAttrs(resultAttrDict);

  // Convert argument types to those required by the binding ABI.
  //
  // NOTE: this is where we could change our signature to provide additional
  // values from the runtime bindings as may be required - like semaphores for
  // async behavior or cancellation.
  auto oldExportType = exportOp.getFunctionType();
  SmallVector<Type> inputTypes;
  for (auto oldType : oldExportType.getInputs()) {
    inputTypes.push_back(mapToABIType(oldType));
  }
  auto fenceType = IREE::HAL::FenceType::get(exportOp.getContext());
  switch (invocationModel) {
    default:
    case IREE::ABI::InvocationModel::Sync:
      break;
    case IREE::ABI::InvocationModel::CoarseFences:
      inputTypes.push_back(fenceType);  // wait
      inputTypes.push_back(fenceType);  // signal
      argAttrDict.push_back(nullptr);   // wait
      argAttrDict.push_back(nullptr);   // signal
      break;
  }
  SmallVector<Type> resultTypes;
  for (auto oldType : oldExportType.getResults()) {
    resultTypes.push_back(mapToABIType(oldType));
  }
  auto newExportType =
      FunctionType::get(exportOp.getContext(), inputTypes, resultTypes);

  // Update the import to the new type and mark it as being converted so we
  // don't try to convert it again.
  auto wrapperOp =
      func::FuncOp::create(exportOp.getLoc(), publicName, newExportType);
  wrapperOp.setPublic();
  wrapperOp->setAttr("iree.abi.stub", UnitAttr::get(exportOp.getContext()));
  wrapperOp.setAllArgAttrs(argAttrDict);
  wrapperOp.setAllResultAttrs(resultAttrDict);

  // Populate the reflection attrs based on the original types.
  populateReflectionAttrs(invocationModel, exportOp, wrapperOp);

  auto *entryBlock = wrapperOp.addEntryBlock();
  auto entryBuilder = OpBuilder::atBlockBegin(entryBlock);

  // Build a map of result value to the argument that has its backing storage.
  SmallVector<Value> resultStorages;
  resultStorages.resize(resultTypes.size());
  for (unsigned i = 0; i < exportOp.getNumArguments(); ++i) {
    auto outputAttr =
        exportOp.getArgAttrOfType<IntegerAttr>(i, "iree.abi.output");
    if (!outputAttr) continue;
    // Today all outputs need to be a !hal.buffer - we could change this
    // in the future to be something more generalized.
    auto storageArg = entryBlock->getArgument(i);
    if (!storageArg.getType().isa<IREE::HAL::BufferType>()) {
      exportOp.emitError() << "storage argument " << i
                           << " has an invalid type " << storageArg.getType()
                           << "; must be a !hal.buffer";
      return {};
    }
    resultStorages[outputAttr.getInt()] = storageArg;
  }

  // Build a map of each I/O argument to the fence that covers them.
  // TODO(benvanik): actually support a map; for now we just handle the 1:M
  // coarse mode where all inputs are covered by a single wait fence and all
  // outputs are covered by a single signal fence.
  Value waitFence;
  Value signalFence;
  switch (invocationModel) {
    default:
    case IREE::ABI::InvocationModel::Sync:
      break;
    case IREE::ABI::InvocationModel::CoarseFences:
      waitFence = entryBlock->getArgument(entryBlock->getNumArguments() - 2);
      signalFence = entryBlock->getArgument(entryBlock->getNumArguments() - 1);
      break;
  }

  // Marshal arguments.
  SmallVector<Value> arguments;
  for (auto arg : llvm::enumerate(
           entryBlock->getArguments().slice(0, oldExportType.getNumInputs()))) {
    auto oldType = oldExportType.getInput(arg.index());
    if (oldType.isa<TensorType>()) {
      auto argLoc = arg.value().getLoc();
      auto importOp = entryBuilder.create<IREE::HAL::TensorImportOp>(
          argLoc, oldType, arg.value(), waitFence);
      arguments.push_back(importOp.getTarget());
    } else {
      arguments.push_back(arg.value());
    }
  }

  // Make the call with the original types.
  auto callOp =
      entryBuilder.create<func::CallOp>(exportOp.getLoc(), exportOp, arguments);
  auto asyncResults = llvm::to_vector(callOp.getResults());

  // Insert a barrier if requested - all tensors will be calculated and the
  // fence will be signaled. Note that even if there are no tensor results we
  // need to signal the fence.
  if (signalFence) {
    SmallVector<Value> asyncTensors;
    for (auto result : asyncResults) {
      if (result.getType().isa<TensorType>()) asyncTensors.push_back(result);
    }
    if (asyncTensors.empty()) {
      // TODO(benvanik): maybe use a global timeline? global stores may not
      // have completed by now in cases where the user wants to loop back.
      entryBuilder.create<IREE::HAL::FenceSignalOp>(exportOp.getLoc(),
                                                    signalFence);
    } else {
      auto barrierOp = entryBuilder.create<IREE::HAL::TensorBarrierOp>(
          exportOp.getLoc(), asyncTensors, signalFence);
      asyncResults = llvm::to_vector(barrierOp.getResults());
    }
  }

  // Marshal results.
  SmallVector<Value> results;
  for (auto [resultIndex, result] : llvm::enumerate(asyncResults)) {
    auto oldType = oldExportType.getResult(resultIndex);
    auto newType = newExportType.getResult(resultIndex);
    if (oldType.isa<TensorType>()) {
      auto dynamicDims = IREE::Util::buildDynamicDimsForValue(
          result.getLoc(), result, entryBuilder);
      results.push_back(entryBuilder.create<IREE::HAL::TensorExportOp>(
          result.getLoc(), newType, result, TypeAttr::get(result.getType()),
          dynamicDims, resultStorages[resultIndex]));
    } else {
      results.push_back(result);
    }
  }

  entryBuilder.create<func::ReturnOp>(exportOp.getLoc(), results);
  return wrapperOp;
}

// Replaces |exportOp| with a wrapper function that exports the native ABI.
// The original function will be made private and be renamed.
// This allows us to support multiple binding schemes as transforms from other
// bindings can also perform their own equivalent wrapping.
static LogicalResult wrapExportFunc(IREE::ABI::InvocationModel invocationModel,
                                    mlir::ModuleOp moduleOp,
                                    func::FuncOp exportOp,
                                    SymbolTable &symbolTable) {
  // Rename the original function so that our wrapper can use the original
  // name in its public definition.
  auto publicName = exportOp.getName().str();
  auto privateName = "_" + publicName;
  auto privateNameAttr =
      mlir::StringAttr::get(exportOp.getContext(), privateName);
  if (failed(symbolTable.replaceAllSymbolUses(exportOp, privateNameAttr,
                                              moduleOp))) {
    return exportOp.emitError() << "unknown symbol table op encountered; "
                                   "cannot fix up symbol names";
  }
  exportOp.setName(privateNameAttr);
  exportOp.setPrivate();

  // Create the wrapper function that conforms to the IREE native ABI and
  // marshals arguments/results to the original function.
  auto wrapperOp =
      createExportWrapperFunc(invocationModel, exportOp, publicName);
  if (!wrapperOp) return failure();
  moduleOp.insert(Block::iterator(exportOp), wrapperOp);

  return success();
}

// Wraps all entry points in a function that is compatible with the
// expected invocation semantics of bindings following the native IREE ABI.
// Imports are also handled as they are entry points in another module.
class WrapEntryPointsPass
    : public PassWrapper<WrapEntryPointsPass, OperationPass<ModuleOp>> {
 public:
  WrapEntryPointsPass() = default;
  WrapEntryPointsPass(const WrapEntryPointsPass &pass) {}
  WrapEntryPointsPass(IREE::ABI::InvocationModel invocationModel) {
    this->invocationModel = invocationModel;
  }

  void getDependentDialects(DialectRegistry &registry) const override {
    registry.insert<func::FuncDialect, mlir::arith::ArithDialect,
                    mlir::tensor::TensorDialect, IREE::HAL::HALDialect>();
  }

  StringRef getArgument() const override {
    return "iree-abi-wrap-entry-points";
  }

  StringRef getDescription() const override {
    return "Wraps all entry points in a function that is compatible with the "
           "expected invocation semantics of bindings following the native "
           "IREE ABI.";
  }

  void runOnOperation() override {
    auto moduleOp = getOperation();

    // Gather functions that need wrapping.
    SmallVector<func::FuncOp> importOps;
    SmallVector<func::FuncOp> exportOps;
    for (auto funcOp : moduleOp.getOps<func::FuncOp>()) {
      // Ignore functions already marked as having their ABI goo handled.
      if (funcOp->hasAttr("iree.abi.stub")) continue;
      if (funcOp.isExternal()) {
        // Imported function.
        importOps.push_back(funcOp);
      } else if (funcOp.isPublic()) {
        // Exported function.
        exportOps.push_back(funcOp);
      }
    }

    SymbolTable symbolTable(moduleOp);

    // Create a wrapper function for each imported function.
    // This will preserve the internal types (tensors/etc) but change the import
    // to taking the ABI types and rewrite calls.
    for (auto importOp : importOps) {
      if (failed(wrapImportFunc(getInvocationModel(importOp, invocationModel),
                                moduleOp, importOp, symbolTable))) {
        return signalPassFailure();
      }
    }

    // Create a wrapper function for each exported function.
    // This will change the export to taking the ABI types and preserve the
    // internal types.
    for (auto exportOp : exportOps) {
      if (failed(wrapExportFunc(getInvocationModel(exportOp, invocationModel),
                                moduleOp, exportOp, symbolTable))) {
        return signalPassFailure();
      }
    }
  }

 private:
  Option<InvocationModel> invocationModel{
      *this,
      "invocation-model",
      llvm::cl::desc("Specifies the execution model used for invocations."),
      llvm::cl::init(IREE::ABI::InvocationModel::Sync),
      llvm::cl::values(
          clEnumValN(IREE::ABI::InvocationModel::Sync, "sync",
                     "Fully synchronous behavior with no fences."),
          clEnumValN(IREE::ABI::InvocationModel::CoarseFences, "coarse-fences",
                     "Exposes one wait fence for all inputs and one signal "
                     "fence for all outputs.")),
  };
};

std::unique_ptr<OperationPass<ModuleOp>> createWrapEntryPointsPass(
    IREE::ABI::InvocationModel invocationModel) {
  return std::make_unique<WrapEntryPointsPass>(invocationModel);
}

static PassRegistration<WrapEntryPointsPass> pass;

}  // namespace ABI
}  // namespace IREE
}  // namespace iree_compiler
}  // namespace mlir
